import sys
import numpy as np
import matplotlib.pyplot as plt
import time
import os

def read_fvecs(filename):
    """Function to read a binary file in 'fvec' format (See https://gforge.inria.fr/projects/yael/).
    
    Input: filename of the fvec file.
    Returns: tuple with descriptor locations and descriptors."""
    pf = open(filename, 'rb')
    # determine size of vectors
    dim = np.fromfile(pf,dtype=np.int32, count=1)[0]
    pf.seek(0, os.SEEK_END)
    size = pf.tell()
    pf.close()
    #print 'Vector dimension: %dx%d'%(size/((dim+1)*4),dim)
    assert dim>0 and dim<10000
    pf = open(filename, 'rb')
    data = np.fromfile(pf, dtype=np.float32, count=size/4)
    data.shape=(size/(4*(dim+1)), dim+1)
    #remove first column, and factor 
    descriptors = data[:,3:]
    uv          = data[:,1:3].astype('int32')
    return uv, descriptors
#-----------------------------------------------------
    
def read_pcd_file(filename):
    """Load a text PCD file. Returns x,y and z matrices."""
    pf = open(filename)
    lines = pf.readlines()
    if lines[1].split()[0]=='VERSION':
        header = lines[0:11]
        width = int(header[6].split()[1])
        height = int(header[7].split()[1])
    else:
        header = lines[0:10]
        width = int(header[5].split()[1])
        height = int(header[6].split()[1])
    x = np.zeros((height, width))
    y = np.zeros((height, width))
    z = np.zeros((height, width))

    i = 0
    j = 0

    fourdims=False
    if len(lines[11].split())==4:
        fourdims=True
    for l in xrange(11, len(lines)):
        if fourdims==True: # maybe some day actually get the color...
            px, py, pz, dump = lines[l].split()
        else:
            px, py, pz = lines[l].split()
        x[j,i] = float(px)
        y[j,i] = float(py)
        z[j,i] = float(pz)
        
        i+=1
        if i>=width:
            j+=1
            i=0
    return x,y,z
#-----------------------------------------------------

def drawbox(plot, u, v, offset):
    """Draws a bounding box in the active figure."""
    plot.plot([u-offset,u+offset],[v+offset,v+offset],'-m')
    plot.plot([u-offset,u-offset],[v-offset,v+offset],'-m')
    plot.plot([u+offset,u+offset],[v-offset,v+offset],'-m')
    plot.plot([u-offset,u+offset],[v-offset,v-offset],'-m')
#-----------------------------------------------------    

def paint_descriptor_xy(npy_desc, centers, plot = plt, fig_num = 1):
    """Creates a figure with a representation of a FINDDD descriptor."""
    numcenters = centers.shape[0]
    nbins=int(np.sqrt(npy_desc.shape[0]/numcenters))
    fig = plot.figure(fig_num)
    fig.clf()
    for i in range(nbins*nbins):
        ax = fig.add_subplot(nbins, nbins, nbins*nbins-i)
        ax.set_xlim([-1.1, 1.1])
        ax.set_ylim([-1.1, 1.1])
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        #centroides
        ax.plot(centers[:,0],centers[:,1],'bo')
        #bins
        subdesc=npy_desc[i*numcenters:(i+1)*numcenters]
        if sum(subdesc)>0:
            subdesc/=sum(subdesc)
        for j,v in enumerate(subdesc):
            #print subdesc
            ax.plot(centers[j,0],centers[j,1], 'ro', markersize=v*20)
#-----------------------------------------------------

def paint_descriptor_and_depth_map_xy(uv, npy_desc, centers):
    """Plot a representation of a FINDDD and mark its location in a depth map."""
    plt.figure(2)
    plt.plot(uv[0], uv[1],'+m', markersize=15)
    drawbox(plt, uv[0], uv[1], 32)
    paint_descriptor_xy(npy_desc, centers, plot = plt, fig_num = 1)
    plt.draw()
    plt.draw()
#-----------------------------------------------------

def paint_descriptor_set(filename_descriptors, filename_pcd, filename_centers, step=10):
    """Display, one by one, all descriptors and their locations in a descriptor set."""
    # get descriptors
    if filename_descriptors.endswith('.geofvecs') or filename_descriptors.endswith('.fvecs'):
        uv, descriptors  = read_fvecs(filename_descriptors)
    else: #txt format
        descriptors  = np.loadtxt(filename_descriptors)
        uv = descriptors[:,:2]
        descriptors = descriptors[:,2:]

    descriptors        = descriptors[::step,:] #show one every <step> descriptors
    uv                 = uv[::step,:]
    
    #get pcd data
    x_pts,y_pts,z_pts = read_pcd_file(filename_pcd)

    #get centers
    centers = np.loadtxt(filename_centers)

    time.sleep(2) # wait for 2 sec the first time
    plt.ion()
    fig=plt.figure(2)

    for ptuv, d in zip(uv, descriptors):
        fig=plt.figure(2)
        fig.clf()
        plt.imshow(z_pts*-1) #flip z-values
        plt.ylim([0, z_pts.shape[0]])
        plt.xlim([0, z_pts.shape[1]])
        paint_descriptor_and_depth_map_xy(ptuv, d, centers)
        plt.draw()
        plt.draw()
        time.sleep(0.1)
        #no nice way to exit(?)

def about():
    print "\nFINDDD Visualization\n--------------------\nTool to visualize descriptors over the Z dimension of a PCD file.\nUsage: python "+sys.argv[0]+" -f <descriptors_file> -pcd <pcd_file> -c [centers_file]\n\nDescription:\nTwo figures display all the information. The first one shows the depth map and a box indicating the location and area of the descriptor, and the second one the histograms of the orientations of the normals, represented as a semi-sphere, seen from the top, for each of the sub-descriptors. NOTE: Does not work with PCA projected descriptors.\n"
    sys.exit(-1)

if __name__=="__main__":
    """Paint, one by one, the descriptors of a fvec file over a depth map.
    Does not work with PCA projected descriptors!"""
    
    filebins = 'xyz_centers13.txt'
    filedesc = None
    filepcd  = None
    #parse args
    args=sys.argv[1:]
    while len(args):
        a=args.pop(0)
        if a   == '-c':       
            filebins = args.pop(0)
        elif a =='-f':         
            filedesc = args.pop(0)
        elif a == '-pcd':     
            filepcd  = args.pop(0)
        else: 
            about()
    if filedesc==None or filepcd==None:
        about()

    paint_descriptor_set(filedesc, filepcd, filebins)
