#include "finddd_descriptor.hpp"


#define PI 3.1415926535

/** 
 * \param nb_bins Number of bins in a sub-descriptor.
 * \return Empty voting vector.
 *
 * Get empty normal voting vector.
 */
inline std::vector<float> get_bin_float_empty(int nb_bins)
{
  std::vector<float> val_bins;
  val_bins.resize(nb_bins,0);
  return val_bins;
}

/** 
 * \param x X coordinate of normal 
 * \param y Y coordinate of normal 
 * \param z Z coordinate of normal 
 * \param bin_centers_x Vector with X coordinate of normal centers
 * \param bin_centers_y Vector with Y coordinate of normal centers
 * \param bin_centers_z Vector with Z coordinate of normal centers
 * \param radius_inf Inter-bin distance among the normal centers.
 * \return Voting vector corresponding to input normal.
 *
 * Get normal voting histogram *with* interpolation.
 */
inline std::vector<float> get_bin_float_interp_xyz(float x, float y, float z, std::vector<float> &bin_centers_x, std::vector<float> &bin_centers_y, std::vector<float> &bin_centers_z, float radius_inf)
{
  int nb_bins=bin_centers_x.size();
  std::vector<float> val_bins;
  val_bins.resize(nb_bins,0);

  for(int ith=0; ith<nb_bins; ith++)
  {
    float dx=(x-bin_centers_x[ith]);
    float dy=(y-bin_centers_y[ith]);
    float dz=(z-bin_centers_z[ith]);
    float dist = sqrt(dx*dx+dy*dy+dz*dz);
    //weight based on one unit of histogram distance
    float vote = std::max(1-dist/radius_inf, (float)0.0);
    val_bins[ith] = vote;
  }
  return val_bins;
}

/** 
 * \param x X coordinate of normal 
 * \param y Y coordinate of normal 
 * \param z Z coordinate of normal 
 * \param bin_centers_x Vector with X coordinate of normal centers
 * \param bin_centers_y Vector with Y coordinate of normal centers
 * \param bin_centers_z Vector with Z coordinate of normal centers
 * \return Voting vector corresponding to input normal.
 *
 * Get normal voting histogram *without* interpolation.
 */
inline std::vector<float> get_bin_float_xyz(float x, float y, float z, std::vector<float> &bin_centers_x, std::vector<float> &bin_centers_y, std::vector<float> &bin_centers_z)
{
  int nb_bins=bin_centers_x.size();
  std::vector<float> val_bins;
  val_bins.resize(nb_bins,0);

  int best_match=0;
  float dist_best_match=999999999999;

  for(int ith=0; ith<nb_bins; ith++)
  {
    float dx=(x-bin_centers_x[ith]);
    float dy=(y-bin_centers_y[ith]);
    float dz=(z-bin_centers_z[ith]);
    float dist = sqrt(dx*dx+dy*dy+dz*dz);
    if(dist < dist_best_match)
    {
      best_match = ith;
      dist_best_match = dist;
    }
  }
  //vote weight = 1
  val_bins[best_match] = 1;

  return val_bins;
}

/**
 * /param pcl_normals Output cloud with 3D point normals computed using PCL library.
 * /param cloud Input cloud with 3D points we want to compute normals for.
 *
 * Function to compute normal vectors fast taking advantage of the fact that the input point cloud is structured. 
 * Compute pointcloud normals fast using integral images.
 */
void compute_normals_integral(pcl::PointCloud<pcl::Normal> &pcl_normals, pcl::PointCloud<pcl::PointXYZ>& cloud)
{
  //compute normals
  pcl::IntegralImageNormalEstimation<pcl::PointXYZ, pcl::Normal> normal_estimator;
  normal_estimator.setNormalEstimationMethod (normal_estimator.AVERAGE_3D_GRADIENT);
  normal_estimator.setMaxDepthChangeFactor(0.02f);
  normal_estimator.setNormalSmoothingSize(10.0f);
  normal_estimator.setInputCloud(boost::make_shared<pcl::PointCloud<pcl::PointXYZ> > (cloud));
  normal_estimator.compute (pcl_normals);
  //For debug: pcl::io::savePCDFileASCII("/tmp/cloud_pointnormals.pcd", pcl_normals);
}

/**
 * /param pcl_normals Output cloud with 3D point normals computed using PCL library.
 * /param cloud Input cloud with 3D points we want to compute normals for.
 *
 * This is a much slower function to compute normal vectors, apt for non-structured point clouds.
 * Compute pointcloud normals *without* using integral images.
 */
void compute_normals(pcl::PointCloud<pcl::Normal> &pcl_normals, pcl::PointCloud<pcl::PointXYZ>& cloud)
{
  //compute normal
  pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normal_estimator;
  normal_estimator.setInputCloud(boost::make_shared<pcl::PointCloud<pcl::PointXYZ> > (cloud));
  normal_estimator.setKSearch (400);
  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ> ());
  normal_estimator.setSearchMethod (tree);
  normal_estimator.compute (pcl_normals);
}

/***************************************************************************************/
/** 
 * \brief Simple Integral Images class.
 * Class to compute simple parallel integral images with floating point accumulators.
 * Includes function to compute sum of sub-rectangle.
 */
class IIf
{
public:
  /** \short Vector of float* to store the parallel integral images in linear form.*/
  std::vector<float*> ii_;
  /** \short Number of bins.*/
  unsigned int bins_;
  /** \short Width of the integral image.*/
  unsigned int width_;
  /** \short Height of the integral image.*/
  unsigned int height_;

  /**
   * \param bins Number of bins/parallel integral images.
   * \param width Width of the integral image.
   * \param height Height of the integral image.
   *
   * Constructor for empty set of parallel integral image.
   */
  IIf(unsigned int bins, unsigned int width, unsigned int height):bins_(bins),width_(width),height_(height){
    ii_.resize(bins);
    for(unsigned int i=0;i<bins;i++)
      ii_[i]=new float[width*height];
  }

  /** Destructor */
  ~IIf(){
    for(unsigned int i=0;i<this->bins_;i++)
      delete [] ii_[i];
  }

  /** 
   * \param im Input vote matrix
   *
   * Function to construct the set of integral images given an input vote matrix.
   */
  void construct(std::vector<std::vector<float> > &im){
    //0,0 case
    unsigned int u=0;
    unsigned int v=0;
    for(unsigned int b=0;b<this->bins_;b++)
    {
      ii_[b][this->width_*v+u] = im[this->width_*v+u][b];
    }
    //n,0 case
    v=0;
    for(u=1;u<this->width_;u++)
    {
      for(unsigned int b=0;b<this->bins_;b++)
      {
	ii_[b][this->width_*v+u] =		     \
	  ii_[b][this->width_*v+(u-1)] +	     \
	  im[this->width_*v+u][b];
      }
    }
    //0,n case
    u=0;
    for(v=1;v<this->height_;v++)
    {
      for(unsigned int b=0;b<this->bins_;b++)
      {
	ii_[b][this->width_*v+u] =		     \
	  ii_[b][this->width_*(v-1)+u] +	     \
	  im[this->width_*v+u][b];
      }
    }
    //n,n case
    for(v=1;v<this->height_;v++)
    {
      for(u=1;u<this->width_;u++)
      {
	for(unsigned int b=0;b<this->bins_;b++)
	{
	  ii_[b][this->width_*v+u] =		     \
	    ii_[b][this->width_*(v-1)+u] +	     \
	    ii_[b][this->width_*v+(u-1)] -	     \
	    ii_[b][this->width_*(v-1)+(u-1)] +	     \
	    im[this->width_*v+u][b];
	}
      }
    }
  }

  /**
   * 
   * \param tu Top-left u coordinate.
   * \param tv Top-left v coordinate.
   * \param bu Bottom-right u coordinate.
   * \param bv Bottom-right v coordinate.
   * \return Vector with sum of the sub-region.
   *
   * Get sum for each parallel integral image for a rectangular sub-region.
   */
  std::vector<float> get_sub_rect(unsigned int tu, unsigned int tv, unsigned int bu, unsigned int bv)
  {
    std::vector<float> ret(this->bins_);
    for(unsigned int b=0;b<this->bins_;b++)
    {
      ret[b] = ii_[b][this->width_*tv+tu] + ii_[b][this->width_*bv+bu] - ii_[b][this->width_*tv+bu] - ii_[b][this->width_*bv+tu];
      if(ret[b]<0){
	ret[b]=0;
      }
    }
    return ret;
  }
};
/***************************************************************************************/

// Constructor
FindddAlgorithm::FindddAlgorithm(void)
{
  //default values
  this->desc_num_side_spatial_bins_ = 1;
  this->desc_patch_radius_=15;  
  
  //patch radius
  this->set_desc_patch_radius(15);

  //Spatial bins (side)
  this->set_num_side_spatial_bins(1);

  //Sample each (num) pixels
  this->sample_each_ = 1;
  this->desc_compute_positionsX_.clear();
  this->desc_compute_positionsY_.clear();
  this->set_positions_file("");
  this->update_pca_matrix("",0);
}

FindddAlgorithm::~FindddAlgorithm(void)
{
}

// Construct with config
FindddAlgorithm::FindddAlgorithm(FindddConfig new_cfg)
{
  this->set_num_side_spatial_bins(new_cfg.num_spatial_bins);
  this->set_desc_patch_radius(new_cfg.desc_patch_radius);
  this->set_sample_each(new_cfg.sample_each);
  this->set_centroids_file(new_cfg.centroids_file);
  this->update_centroids();
  this->set_positions_file(new_cfg.positions_file);
  this->normalize_desc_ = new_cfg.normalize_desc;  // * 0 - No norm // * 1 - L1 // * 2 - L2 // * 3 - sqrt + L2 
  this->use_soft_voting_ = new_cfg.use_soft_voting;
  this->verbose_level_ = new_cfg.verbose_level;
  this->update_pca_matrix(new_cfg.pca_file, new_cfg.pca_nkeep);
}

// Main function to compute FINDDD descriptors
void FindddAlgorithm::compute_ndescs_integral_spatial_interpolation(pcl::PointCloud<pcl::PointXYZ>& cloud, DescriptorSet &descriptor_set)
{
  clock_t start=clock();
  if(this->verbose_level_>=2)
  {
    std::cout<<"II with interpolation method."<<std::endl;
    std::cout<<"PointCloud received"<<std::endl;
  }
  //compute normals
  pcl::PointCloud<pcl::Normal> pcl_normals; 
  compute_normals_integral(pcl_normals, cloud);

  //set up object to hold descriptors
  descriptor_set.descriptors.clear();
  descriptor_set.len    = this->desc_num_total_bins_;
  descriptor_set.width  = pcl_normals.width;
  descriptor_set.height = pcl_normals.height;
  descriptor_set.num_orient_bins = this->orient_bin_centers_x_.size();
  descriptor_set.num_spa_bins = this->desc_num_side_spatial_bins_;

  //Init II
  std::vector<std::vector<float> > binned(cloud.width*cloud.height);

  //Compile all normal orientations and add them to the II
  //soft-voting vs not-soft-voting checked here for efficiency to avoid if in a loop.
  //(sorry for the ugly code).
  if(this->use_soft_voting_) {
    for(uint v=0;v<cloud.height;v++) {
      for(uint u=0;u<cloud.width;u++)  {
	if(isnan(pcl_normals(u,v).normal_x) || isnan(pcl_normals(u,v).normal_y) || isnan(pcl_normals(u,v).normal_z))
	  binned[u+cloud.width*v]=get_bin_float_empty(this->orient_bin_centers_x_.size()); // empty vector
	else {
	  binned[u+cloud.width*v] = get_bin_float_interp_xyz(pcl_normals(u,v).normal_x, pcl_normals(u,v).normal_y, pcl_normals(u,v).normal_z, this->orient_bin_centers_x_, this->orient_bin_centers_y_, this->orient_bin_centers_z_, this->radius_inf_);
	} } } } 
  else {
    for(uint v=0;v<cloud.height;v++) {
      for(uint u=0;u<cloud.width;u++) {
	if(isnan(pcl_normals(u,v).normal_x) || isnan(pcl_normals(u,v).normal_y) || isnan(pcl_normals(u,v).normal_z))
	  binned[u+cloud.width*v] = get_bin_float_empty(this->orient_bin_centers_x_.size()); // empty vector
	else {
	  binned[u+cloud.width*v] = get_bin_float_xyz(pcl_normals(u,v).normal_x, pcl_normals(u,v).normal_y, pcl_normals(u,v).normal_z, this->orient_bin_centers_x_, this->orient_bin_centers_y_, this->orient_bin_centers_z_);
	} } }
  }

  //Build II
  IIf intim(this->desc_num_orient_bins_, cloud.width, cloud.height);
  intim.construct(binned);
  if(this->verbose_level_>=2)
    std::cout<<"built II"<<std::endl;

  unsigned int spatial_bin_size    = (2*this->desc_patch_radius_+1)/this->desc_num_side_spatial_bins_;
  unsigned int spatial_bin_orients = this->desc_num_orient_bins_;

  //If no positions file given, generate standard positions
  if(this->desc_compute_positionsX_.size()==0)
  {
    if(this->verbose_level_>=2)
      std::cout<<"Generating default positions"<<std::endl;
    for(unsigned int v = this->desc_patch_radius_;v<cloud.height-this->desc_patch_radius_-1;v+=this->sample_each_)
    {
      for(unsigned int u = this->desc_patch_radius_;u<cloud.width-this->desc_patch_radius_-1;u+=this->sample_each_)
      {      
	this->desc_compute_positionsX_.push_back(u);
	this->desc_compute_positionsY_.push_back(v);
      }
    }
  }
  
  //Compute descriptors
  for(unsigned int i_pos=0; i_pos<this->desc_compute_positionsX_.size();i_pos++)
  {
    unsigned int u=this->desc_compute_positionsX_[i_pos];
    unsigned int v=this->desc_compute_positionsY_[i_pos];

    //skip invalid margin positions
    if( v<this->desc_patch_radius_ || 
	u<this->desc_patch_radius_ || 
        u>=cloud.width-this->desc_patch_radius_-1 || 
        v>=cloud.height-this->desc_patch_radius_-1)
    {
      continue;
    }
    Descriptor desc;
    desc.point3d.x=cloud(u,v).x;
    desc.point3d.y=cloud(u,v).y;
    desc.point3d.z=cloud(u,v).z;
    
    //insert point2D
    desc.u=u;
    desc.v=v;
    
    desc.descriptor.resize(this->desc_num_total_bins_,0);
    //float total_votes=0; // for normalization
    
    unsigned int absolute_top  = v-this->desc_patch_radius_;
    //int absolute_bottom = v+this->desc_patch_radius_;
    unsigned int absolute_left = u-this->desc_patch_radius_;
    //int absolute_right = u+this->desc_patch_radius_;
    
    for(unsigned int k=0;k<this->desc_num_side_spatial_bins_; k++)
    {
      unsigned int bound_left  = absolute_left + k*spatial_bin_size;
      unsigned int bound_right = absolute_left + (k+1)*spatial_bin_size;
      for(unsigned int l=0; l<this->desc_num_side_spatial_bins_; l++)
      {
	unsigned int bound_top    = absolute_top + l*spatial_bin_size;
	unsigned int bound_bottom = absolute_top + (l+1)*spatial_bin_size;

	std::vector<float> tmp = intim.get_sub_rect(bound_left, bound_top, bound_right, bound_bottom);
	for(unsigned int i=0;i<spatial_bin_orients;i++)
	{
	  unsigned int offset_full_rows = l*this->desc_num_side_spatial_bins_*spatial_bin_orients;
	  unsigned int offset_this_row  = k*spatial_bin_orients;
	  desc.descriptor[offset_full_rows + offset_this_row + i]=tmp[i];
	  //total_votes+=tmp[i];
	}
      }
    }
 
    //normalize desc if required
    if(this->normalize_desc_ != 0)
    {
      float sum=0;
      for(int iii=0; iii<desc.descriptor.size(); iii++)
      {
	if((this->normalize_desc_==1) || (this->normalize_desc_==3)) 
	  sum += desc.descriptor[iii];
	else if(this->normalize_desc_==2) 
	  sum += desc.descriptor[iii]*desc.descriptor[iii];
      }
      if(this->normalize_desc_==2) 
	sum=sqrt(sum);
      for(int iii=0; iii<desc.descriptor.size(); iii++)
      {
	if((this->normalize_desc_==1) || (this->normalize_desc_==2)) 
	  desc.descriptor[iii] /= sum;
	else if(this->normalize_desc_==3) 
	  desc.descriptor[iii] = sqrt(desc.descriptor[iii])/sqrt(sum);
      }
    }
    //project w\ PCA if required
    if(this->pca_file_!="")
    {
      desc.descriptor = this->project_vector_pca(desc.descriptor, this->pca_matrix_, this->pca_mean_);
      if(this->verbose_level_>=2)
	std::cout<<"Dimension of vector "<<desc.descriptor.size()<<" after projecting with PCA"<<std::endl;
    }
    descriptor_set.descriptors.push_back(desc);
    descriptor_set.num++;
  }
  if(this->verbose_level_>=1)
    std::cout<<"Computed "<<descriptor_set.num<<" descriptors, in "<<(clock()-start)/(float)CLOCKS_PER_SEC<< "s"<<std::endl;
}

/** Method parameters help.
 */
void about()
{
  std::cout<<"Usage:\n  finddd_descriptor -pcd <file_path> -centroids_file <centroids_path> -descfile <output_descfile_path> [-use_soft_voting <bool>] [-sample_each <step>] [-normalize_desc] [-desc_patch_radius <radius>] [-num_spatial_bins <num_bins_per_side>] [-positions_file <positions_file_path>] [-keep_nan] [-pca_file <file_path>] [-pca_nkeep <num ev to keep>] [-ascii]"<<std::endl;
  std::cout<<"Parameters:\n  -pcd <filename>              Input pcd file path.\n  -use_soft_voting <bool>      Disable soft-voting.\n  -sample_each <num_pix>       Density of descriptor sampling.\n  -normalize_desc <norm>       0=none 1=L1 2=L2 3=sqrt+L2.\n  -desc_patch_radius <num_pix> Radius of the local patch.\n  -num_spatial_bins <num_bins> Number of local bins per side (sqrt of total bins).\n  -centroids_file <filename>   Input centroid file path.\n  -positions_file <filename>   Input locations file path.\n  -descfile <filename>         Output descriptor file path.\n  -pca_file <filename>         PCA matrix filename\n  -pca_nkeep <num_ev_to_keep>  Number of eigenvectors to keep (output dimension).\n  -ascii                       Save descriptors in ASCII format.\n";
}

/** 
 * \param argc Standard argc, number of arguments.
 * \param argv Standard argv, arguments in vector of char pointers.
 * \param new_conf Output configuration object with parameters according to arguments.
 * \return Boolean, true if all parameters valid.
 *
 * Function to parse command-line arguments.
 */
bool parse_args(int argc, char* argv[], FindddConfig& new_conf)
{
  bool valid_args=true;
  for(int i=1; i<argc; i++)
  {
    std::string arg(argv[i]);
    if(arg == "-pcd") {
      new_conf.pcd_file = std::string(argv[++i]);
    } else if(arg == "-use_soft_voting") {
      new_conf.use_soft_voting = atoi(argv[++i]);
    } else if(arg == "-sample_each") {
      new_conf.sample_each = atoi(argv[++i]);
    } else if(arg == "-normalize_desc") {
      new_conf.normalize_desc = atoi(argv[++i]);
    } else if(arg == "-desc_patch_radius") {
      new_conf.desc_patch_radius = atoi(argv[++i]);
    } else if(arg == "-num_spatial_bins") {
      new_conf.num_spatial_bins = atoi(argv[++i]);
    } else if(arg == "-centroids_file") {
      new_conf.centroids_file = std::string(argv[++i]);
    } else if(arg == "-pca_file") {
      new_conf.pca_file = std::string(argv[++i]);
    } else if(arg == "-pca_nkeep") {
      new_conf.pca_nkeep = atoi(argv[++i]);
    } else if(arg == "-positions_file") {
      new_conf.positions_file = std::string(argv[++i]);
    } else if(arg == "-descfile") {
      new_conf.geofvecs_file = std::string(argv[++i]);
    } else if(arg == "-keep_nan") {
      new_conf.keep_nan_points = 1;
    } else if(arg == "-ascii") {
      new_conf.save_in_ascii = 1;
    } else {
      std::cerr<<"Unknown argument: "<<arg<<"."<<std::endl;
      return false;
    }
  }
  if(new_conf.pcd_file=="")
  {
    std::cerr<<"Missing input pcd file."<<std::endl;
    valid_args=false;
  }
  if (new_conf.centroids_file=="")
  {
    std::cerr<<"Missing centroids file."<<std::endl;
    valid_args=false;
  }
  if(new_conf.geofvecs_file=="")
  {
    std::cerr<<"Missing output descriptorsfile."<<std::endl;
    valid_args=false;    
  }
  if(valid_args==true)
    return true;
  else
    return false;
}

/** 
 * \param pf File pointer where to save the data.
 * \param point Descriptor to be saved.
 *
 * Function to write a descriptor to a file pointer in fvec format (see the yael library for more details: https://gforge.inria.fr/projects/yael/)
 * Save descs in fvec format.
 */
void write_point_ascii(FILE *pf, const Descriptor &point)
{
  int dim=point.descriptor.size();
  if (fprintf(pf, "%d %d %d ", point.u, point.v, dim)<0)
  {
    fprintf(stderr, "Error writting to file\n"); exit(-1);
  }
  for(unsigned int j=0;j<dim-2;j++)
    if (fprintf(pf, "%f ", point.descriptor[j])<0)
    {
      fprintf(stderr, "Error writting to file\n"); exit(-1);
    }
  fprintf(pf, "\n");
}

/** 
 * \param pf File pointer where to save the data.
 * \param point Descriptor to be saved.
 *
 * Function to write a descriptor to a file pointer in fvec format (see the yael library for more details: https://gforge.inria.fr/projects/yael/)
 * Save descs in fvec format.
 */
void write_point_fvec(FILE *pf, const Descriptor &point)
{
  int dim=point.descriptor.size();
  float f;
  dim+=2; //adding x,y coord
  if (fwrite (&(dim), sizeof (int), 1, pf) != 1)
  {
    fprintf(stderr, "Error writting to file\n"); exit(-1);
  }
  f=point.u;
  if (fwrite (&f, sizeof (float), 1, pf) != 1)
  {
    fprintf(stderr, "Error writting to file\n"); exit(-1);
  }
  f=point.v;
  if (fwrite (&f, sizeof (float), 1, pf) != 1)
  {
    fprintf(stderr, "Error writting to file\n"); exit(-1);
  }
  for(unsigned int j=0;j<dim-2;j++)
    if (fwrite (&(point.descriptor[j]), sizeof (float), 1, pf) != 1)
    {
      fprintf(stderr, "Error writting to file\n"); exit(-1);
    }
}

/** 
 * \param output_file Filename where descriptors will be saved.
 * \param descriptor_set Descriptor set to be saved.
 * \param conf Configureation parameters of the descriptors.
 * Function to save to disc a set of descriptors computed in a range image. The conf parameter is used to check if descriptors corresponding to NaN 3D points have to be saved or discarded.
 * Save valid descriptors to disk.
 */
void write_feats(std::string &output_file, DescriptorSet &descriptor_set, FindddConfig &conf)
{
  FILE* pf=0;
  if(conf.save_in_ascii)
    pf=fopen(output_file.c_str(),"w");
  else
    pf=fopen(output_file.c_str(),"wb");
  int saved_points = 0;

  for(int ii=0;ii<descriptor_set.num;ii++)
  {
    if( (!isnan(descriptor_set.descriptors[ii].point3d.x) &&
	 !isnan(descriptor_set.descriptors[ii].point3d.y) &&
	 !isnan(descriptor_set.descriptors[ii].point3d.z)) ||
	conf.keep_nan_points)
    {
      if(conf.save_in_ascii) write_point_ascii(pf, descriptor_set.descriptors[ii]);
      else write_point_fvec(pf, descriptor_set.descriptors[ii]);
      saved_points++;
    }
  }
  fclose(pf);
}

int main(int argc, char* argv[])
{
  // parse arguments
  FindddConfig new_conf;
  bool args_ok = false;
  args_ok = parse_args(argc, argv, new_conf);
  // For debug:
  // std::cout<<"pcdfile="<<new_conf.pcd_file<<" centroids="<<new_conf.centroids_file<<" positions="<<new_conf.positions_file<<" geofvecsfile="<<new_conf.geofvecs_file<<" num_spatial_bins="<<new_conf.num_spatial_bins<<" desc_patch_radius="<<new_conf.desc_patch_radius<<" sample_each="<<new_conf.sample_each<<" normalize_desc="<<new_conf.normalize_desc<<" use_soft_voting="<<new_conf.use_soft_voting<<" keep_nan_points="<<new_conf.keep_nan_points<<std::endl;

  if(args_ok)
  {
    // create finddd object
    FindddAlgorithm comp(new_conf);
    // load pointcloud
    pcl::PointCloud<pcl::PointXYZ> cloud;
    if(pcl::io::loadPCDFile<pcl::PointXYZ>(new_conf.pcd_file.c_str(), cloud)==-1)
    {
      std::cerr<<"Error in pcd file: "<<new_conf.pcd_file<<std::endl;
      exit(-1);
    }
    DescriptorSet descriptor_set;
    // compute features
    comp.compute_ndescs_integral_spatial_interpolation(cloud, descriptor_set);
    // save features to disk
    write_feats(new_conf.geofvecs_file, descriptor_set, new_conf);
  }
  else
  {
    about();
  }
  
  return 0;
}





