
#ifndef _finddd_alg_h_
#define _finddd_alg_h_

#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
//#include <pcl/ros/conversions.h>
#include <pcl/features/normal_3d.h>
#include <pcl/features/integral_image_normal.h>
#include <pcl/io/pcd_io.h>
#include <boost/make_shared.hpp>
#include <cmath>
#include <vector>
#include <sstream>
#include <time.h>
#include <iostream>
#include <fstream>

// util
template <class T>
inline std::string to_string (const T& t)
{
  std::stringstream ss;
  ss << t;
  return ss.str();
}

/** \brief Class for 3D point.
 *  Very simple class to hold 3D point position information.
 */
class P3D
{
public:
  float x;
  float y;
  float z;
  
  P3D():x(0),y(0),z(0) {}
  P3D(float xx, float yy, float zz):x(xx),y(yy),z(zz) {}
  
};

/** \brief Class to hold one FINDDD descriptor.
 *  This class holds each one of the FINDDD descriptors in a set. 
 *  %Descriptor data is stored in a vector<float>.
 */
class Descriptor
{
public:
  /** \short X-coordinate of the descriptor in image. */
  int u;
  /** \short Y-coordinate of the descriptor in image. */  
  int v;
  /** \short Orientation of the descriptor in image (not used yet). */
  float orientation;  
  /** \short Corresponding 3D coordinates of the descriptor. */
  P3D point3d; 
  /** \short %Descriptor vector. */
  std::vector<float> descriptor;
};

/** \brief Class to hold FINDDD computation parameters.
 *  This class holds a set (vector) of FINDDD descriptors as well as 
 *  parameters of the descriptors in the set and sizes of the original
 *  range image.
 */
class DescriptorSet
{
public:
  /** \short Number of orientation bins of each sub-descriptor. */
  int num_orient_bins;
  /** \short Number of spatial bins per side of the full descriptor region (sqrt of total). */
  int num_spa_bins;
  /** \short Number of descriptors. */
  int num;
  /** \short Number of dimensions of the descriptor. */
  int len;
  /** \short Width of the original range image. */
  int width;
  /** \short Height of the original range image. */
  int height;
  /** \short Vector with computed descriptor. */
  std::vector<Descriptor> descriptors;
  
  DescriptorSet()
  {
    num_orient_bins=0;
    num_spa_bins=0;
    num=0;
    len=0;
    width=0;
    height=0;
  }
};


/** \brief Class to hold FINDDD computation parameters.
 *  This class holds all parameters related to FINDDD computation, and is 
 *  used to update the FindddAlgorithm class with user-defined parameters.
 */
class FindddConfig
{
public:
  /** \short Num of spatial subdivisions. */
  int num_spatial_bins;
  /** \short Side of the patch in pixels. */
  int desc_patch_radius;
  /** \short Step in pixels between descriptors. */
  int sample_each;
  /** \short File with the normal centroids. */
  std::string centroids_file;
  /** \short File with the (u,v) positions where descriptors have to be computed (overrides "sample_each"). */
  std::string positions_file;
  /** \short Normalization of new descriptors.

      Normalization of new descriptors:
      - 0 - No norm
      - 1 - L1
      - 2 - L2
      - 3 - sqrt + L2 
  */
  int normalize_desc;
  /** \short Whether to use soft-voting or direct assignment. */
  int use_soft_voting;
  /** \short Input pcd file where descriptors have to be computed. */
  std::string pcd_file;
  /** \short Output file in "fvec" format containing the computed descriptors. */
  std::string geofvecs_file;
  /** \short Whether to keep NaN points or discard them. */
  int keep_nan_points;
  /** \short File with PCA matrix.

      Name of a file containing the mean and the PCA matrix to automatically compress the computed descriptors. If not specified no compression is done. */
  std::string pca_file;
  /** \short Number of eigenvectors to keep in the PCA matrix (=0 to keep all of them). */
  int pca_nkeep;
  /** \short Verbosity level. */
  int verbose_level;
  /** \short save descriptors in ascii format */
  int save_in_ascii;
  FindddConfig()
  {
    verbose_level=1;
    num_spatial_bins=4;
    desc_patch_radius=21;
    sample_each=6;
    centroids_file="";
    positions_file="";
    normalize_desc=0;
    use_soft_voting=1;
    pcd_file="";
    geofvecs_file="";
    keep_nan_points=0;
    pca_file="";
    pca_nkeep=0;
    save_in_ascii=0;
  }
};

/** \brief Main class to compute FINDDD descriptors.
 * This class contains all the methods necessary to compute FINDDD
 * descriptors for a range image.
 */
class FindddAlgorithm
{
protected:

  // private attributes and methods
  uint nt_;
  
  /** \short Vector containing the X coordinate of normal centroids. */
  std::vector<float> orient_bin_centers_x_;
  /** \short Vector containing the Y coordinate of normal centroids. */
  std::vector<float> orient_bin_centers_y_;
  /** \short Vector containing the Z coordinate of normal centroids. */
  std::vector<float> orient_bin_centers_z_;
  
  /** \short Vector containing the U coordinate of points where descriptors have to be computed. */
  std::vector<int> desc_compute_positionsX_;
  /** \short Vector containing the V coordinate of points where descriptors have to be computed. */
  std::vector<int> desc_compute_positionsY_;
  /** \short Inter-bin distance, to compute soft-votes*/
  float radius_inf_;
  /** \short Number of orientation bins in each sub-descriptor. */
  uint desc_num_orient_bins_;
  /** \short Number of spatial bins per side of the full descriptor region (sqrt of total). */
  uint desc_num_side_spatial_bins_;
  /** \short Total size of the descriptors. */
  uint desc_num_total_bins_;
  /** \short Radius of area used to compute descriptor. */
  uint desc_patch_radius_;
  /** \short Filename of the normal centroids. */
  std::string centroids_file_;
  /** \short File with positions to compute descriptors.

      Filename of the file with positions where descriptors have to be computed (if specidied overrides "sample_each"). */
  std::string positions_file_;
  /** \short File with PCA matrix

      Name of file with mean and PCA matrix to compress descriptors on the fly. If not specified no compression is done. */
  std::string pca_file_;
  /** \short Number of eigenvectors to keep in the PCA matrix (=0 to keep all of them)*/
  int pca_nkeep_;
  /** \short Normalization of new descriptors.

      Normalization of new descriptors:
      - 0 - No norm
      - 1 - L1
      - 2 - L2
      - 3 - sqrt + L2 
  */
  uint normalize_desc_;
  /** \short Whether to use soft-voting or direct assignment. */
  uint use_soft_voting_;
  /** \short Level of verbosity. */
  int verbose_level_;
  /** \short Step in pixels between descriptors. */
  uint sample_each_;
  /** \short Margin to skip around the image. Currently this variable is unused. */
  uint margin_;

  // [PCA stuff]
  /** \short Variable holding the PCA matrix for on the fly compression. */
  std::vector< std::vector<float> > pca_matrix_;
  /** \short Mean of descriptors for PCA compression. */
  std::vector<float> pca_mean_;

  public:
  FindddAlgorithm();
  /**
   * \brief constructor.
   *
   * Constructor that takes a FindddConfig object as input parameter.
   *
   */
  FindddAlgorithm(FindddConfig conf);

  /**
   * \brief Set total descriptor bins.
   * 
   * Given the current parameters, this function computes the total
   * number of bins of the descriptor.
   *
   */
  inline int compute_total_bins()
  {
    this->desc_num_total_bins_ = this->desc_num_side_spatial_bins_ * this->desc_num_side_spatial_bins_ * this->orient_bin_centers_x_.size();
    return this->desc_num_total_bins_;
  }
  
  /**
   * \brief Set number of spatial bins (side).
   * \param nb Number of spatial bins per side of the descriptor support area.
   * \return New number of spatial bins per size (same as input). 
   * This function allows to change the number of spatial bins of
   * the descriptor, and triggers all actions dependant on them.
   *
   */
  inline int set_num_side_spatial_bins(int nb)
  {
    this->desc_num_side_spatial_bins_ = nb;
    this->compute_total_bins();
    return this->desc_num_side_spatial_bins_;
  }
  

  /**
   * \brief Set patch radius.
   * \param r Radius of the descriptor support region.
   * \return New radius of descriptor support region (same as input).
   * This function allows to change the patch radius of the
   * descriptor, and triggers all actions dependant on them.
   *
   */
  inline int set_desc_patch_radius(int r)
  {
    this->desc_patch_radius_ = r;
    this->margin_ = this->desc_patch_radius_;
    return this->desc_patch_radius_;
  }

  /**
   * \brief Set sample each
   * \param s Step at which descriptors have to be computed.
   * \return New step at which descriptors have to be computed (same as input).
   * Function used to change the spacing between each sampled
   * position of the grid for descriptor computation.
   *
   */
  inline int set_sample_each(int s)
  {
    this->sample_each_ = s;
    return this->sample_each_;
  }
  
  /**
   * \brief Update PCA matrix.
   * \param new_pca_matrix_file File with the PCA matrix and mean vector.
   * \param pca_nkeep Number of ev that must be kept.
   * Update the PCA matrix used to project the computed 
   * descriptors. The function expects a string pointing
   * to a plain text file with the mean vector followed by
   * the PCA matrix.
   *
   */
  inline void update_pca_matrix(std::string new_pca_matrix_file, int pca_nkeep)
  {
    this->pca_file_ = new_pca_matrix_file;
    this->pca_nkeep_=pca_nkeep;
    if(new_pca_matrix_file=="")
      return;
    std::ifstream ifs(this->pca_file_.c_str(), std::ifstream::in);
    if(!ifs.is_open())
    {
      if(this->verbose_level_>=1) 
	std::cout<<"FINDDD: could not open PCA file. "<<this->pca_file_<<std::endl;
      exit(-1);
    }
    this->pca_matrix_.clear();
    this->pca_mean_.clear();
    
    std::string line;
    if (ifs.good()) // First the mean!
    {
      float num;
      getline(ifs, line);
      std::istringstream iss(line);
      int i=0;
      do 
      {
        iss >> num;
        this->pca_mean_.push_back(num);
      }while(!iss.eof());
    }
    int dout=0;
    while (ifs.good())
    {
      float num;
      getline(ifs, line);
      //line.erase(line.find_last_not_of(" \n\r\t")+1);
      if(line.size()>0)
      {
	this->pca_matrix_.push_back(std::vector<float>());
	std::istringstream iss(line);
	do 
	{
	  iss >> num;
	  this->pca_matrix_.back().push_back(num);
	}while(!iss.eof());
	if(dout==0) dout=this->pca_matrix_.back().size();
	else if(dout!=this->pca_matrix_.back().size())
	{
	  if(this->verbose_level_>=1)
	    std::cout<<"Not consistent PCA data. Previous PCA matrix row width: "<<dout<<". Current PCA matrix row width: "<<this->pca_matrix_.back().size()<<std::endl;
	  exit(-1);
	}
      }
    }
    if(this->pca_mean_.size() != this->pca_matrix_.size())
    {
      if (this->verbose_level_>=1)
	std::cout<<"Not consistent PCA data: mean vector size: "<<this->pca_mean_.size()<<" PCA matrix size: "<<this->pca_matrix_.size()<<"x"<<pca_matrix_[0].size()<<std::endl;
      exit(-1);
    }
    //matrix correctly loaded, keeping only relevant eigenvectors and transposing for ease of use later
    int nkeep=0;
    if(pca_nkeep>0)
      nkeep=pca_nkeep>this->pca_matrix_.size()?this->pca_matrix_.size():pca_nkeep; //min
    else
      nkeep=this->pca_matrix_.size();
    std::vector< std::vector<float> > PCA_t;
    PCA_t.resize(nkeep, std::vector<float>(this->pca_matrix_.size()));
    for(int i=0;i<nkeep;i++)
      for(int j=0;j<this->pca_matrix_.size();j++)
      {
	PCA_t[i][j]=this->pca_matrix_[j][i];
      }
    this->pca_matrix_=PCA_t;
  }

  /**
   * \brief Project descriptor using PCA.
   * \param v Descriptor to compress. 
   * \param PCA PCA matrix.
   * \param mean Mean vector to center descriptor vector.
   * \return Compressed descriptor vector.
   * Project a newly computed descriptor using the PCA
   * matrix passed as parameter. This function is automatically
   * executed if "pca_file_" is not NULL.
   *
   */
  std:: vector<float> project_vector_pca(std::vector<float> &v, std::vector< std::vector<float> > &PCA, std::vector<float> &mean)
  {
    std::vector<float> o;
    o.resize(PCA.size(),0);
    //PCA matrix is transposed, so can be used directly with an iterator
    std::vector<float>::iterator o_it = o.begin();
    for(std::vector< std::vector<float> >::iterator PCA_it=PCA.begin(); PCA_it!=PCA.end(); PCA_it++)
    {
      for(std::vector<float>::iterator v_it=v.begin(), pcav_it=PCA_it->begin(), m_it=mean.begin(); v_it!=v.end(); v_it++, pcav_it++, m_it++)
      {
	*o_it += (*v_it-*m_it) * *pcav_it;
      }
      o_it++;
    }
    return o;
  }

  /**
   * \brief Set centroids file.
   * \param new_centroids_file File containing the normal centroids.
   * Function used to load the bin centers from a file
   *
   */
  inline void set_centroids_file(std::string new_centroids_file)
  {
    this->centroids_file_ = new_centroids_file;
    this->update_centroids();
  }

  /**
   * \brief Load set of normal centers.
   * 
   * Update centroids with contents of the file pointed by the centroids_file_ variable.
   * TODO Not much checking is done to guarantee that the file is consistent!
   *
   */
  inline void update_centroids()
  {
    std::string path = this->centroids_file_.c_str();
    std::ifstream ifs(path.c_str(), std::ifstream::in);
    if(!ifs.is_open())
    {
      if(this->verbose_level_>=1) 
	std::cout<<"FINDDD: Incorrect number of centers loaded from file "<<path<<std::endl;
      exit(-1);
    }

    this->orient_bin_centers_x_.clear();
    this->orient_bin_centers_y_.clear();
    this->orient_bin_centers_z_.clear();
    
    while (ifs.good()) //trusting the file...
    {
      float x,y,z;
      ifs >> x;
      ifs >> y;
      ifs >> z;
      if(ifs.good())
      {
        this->orient_bin_centers_x_.push_back(x);
        this->orient_bin_centers_y_.push_back(y);
        this->orient_bin_centers_z_.push_back(z);
      }
    }
    ifs.close();
    this->desc_num_orient_bins_ = this->orient_bin_centers_x_.size();
    
    if(this->verbose_level_>=2)
      std::cout<<"Loaded "<<this->desc_num_orient_bins_<<" centers"<<std::endl;
    this->compute_total_bins();
    this->compute_radius_inf();
  }
  
  /**
   * \brief Set positions file.
   * \param new_positions_file
   * Function used to load the positions of the image where descriptors
   * will be computed from a list in a file.
   * TODO Not much checking is done to guarantee that the file is consistent!
   *
   */
  inline void set_positions_file(std::string new_positions_file)
  {
    this->positions_file_ = new_positions_file;
    
    this->desc_compute_positionsX_.clear();
    this->desc_compute_positionsY_.clear();
    
    if(this->positions_file_=="") // No positions file specified.
    {
      if(this->verbose_level_>=2)
	std::cout<<"Generating positions with 'sample_each' param."<<std::endl;
      return;
    }
    
    std::string path = this->positions_file_.c_str();
    std::ifstream ifs(path.c_str(), std::ifstream::in); 
    if(!ifs.is_open())
    {
      if(this->verbose_level_>=2)
	std::cout<<"FINDDD: Incorrect descriptor positions file "<<path;
      exit(-1);
    }
    
    while (ifs.good()) //trusting the file...
    {
      int x,y;
      ifs >> x;
      ifs >> y;
      this->desc_compute_positionsX_.push_back(x);
      this->desc_compute_positionsY_.push_back(y);
    }
    ifs.close();
    if(this->verbose_level_>=2)
      std::cout<<"Loaded "<<this->desc_compute_positionsX_.size()<<" positions"<<std::endl;
  }
  
  /**
   * \brief Determine vote influence cutoff.
   *
   * Function to compute minimum distance between bin 0 and the rest of bins, 
   * for vote interpolation. 
   */    
  inline void compute_radius_inf()
  {
    //finding minimum distance between point zero and the other points
    float x = this->orient_bin_centers_x_[0];
    float y = this->orient_bin_centers_y_[0];
    float z = this->orient_bin_centers_z_[0];
    int npoints = this->orient_bin_centers_x_.size();
    this->radius_inf_ = 9999999999;
    for(int i=1; i<npoints; i++)
    {
// Maybe radius_inf could be squared to avoid the sqrt
      float dist = sqrt(pow(x-this->orient_bin_centers_x_[i],2) +
			pow(y-this->orient_bin_centers_y_[i],2) +
			pow(z-this->orient_bin_centers_z_[i],2));
      if(dist<this->radius_inf_)
	this->radius_inf_ = dist;
    }
    if(this->verbose_level_>=2) 
      std::cout<<"Using ratio_inf_ = "<<this->radius_inf_<<std::endl;
  }
  
  /**
   * \brief compute all descriptors with spatial bin support of image using an integral image.
   * \param cloud input structured point cloud loaded from pcd file.
   * \param descriptor_set Output computed descriptors.
   * Main interface to compute normal descriptors with spatial support from a point cloud
   */
  void compute_ndescs_integral_spatial_interpolation(pcl::PointCloud<pcl::PointXYZ>& cloud, DescriptorSet &descriptor_set);
  
  /**
   * \brief Destructor.
   *
   * This destructor is called when the object is about to be destroyed.
   *
   */
  ~FindddAlgorithm(void);
};

#endif
